##
## ODB2 - Pure Python object database
## Copyright (C) 2004 Sami Hangaslammi (shang@jyu.fi)
##
"""
File Storage Utilities
"""

from __future__ import generators

from types import SliceType
from os.path import getsize as _filesize, exists
from bisect import bisect
import math
from sys import maxint
import cPickle
import struct
import time

_BLOCKSIZE = 64*1024
_MAXSIZE = maxint

class allocerror(Exception): pass

def _fmtr(addr,size):
    end = addr+size
    #return "0x%X-0x%X" % (addr,end)
    return "%i-%i" % (addr,end)


## ----- Allocation Table ----- ##

class AllocTab(object):
    """
    AllocTab is used for keeping track which segments of a value
    area are "in use" and which are "free".
    """
    
    def __init__(self):
        self._allocated = []
        self._free = []

    def allocate(self, addr, size):
        """allocate(addr, size) -> None

        Mark 'size' values starting from 'addr' as used. The marked
        values must be free and the allocation should extend a continuous
        "used" segment if possible. Use the find_free method to find
        a valid location for the allocation.
        """
        if size == 0:
            return
        if addr < 0:
            raise allocerror, "invalid allocation address %i" % addr
        if size < 0:
            raise allocerror, "invalid allocation size %i" % size
        free = self._free
        allocated = self._allocated
        
        if not free: # special case: no free space
            if not allocated: # special case: no space or allocations
                if addr != 0:
                    raise allocerror, "invalid allocation %s, must begin allocation at 0x%X" % (_fmtr(size,addr),0)
                allocated.append((addr,size))
                return
            if len(allocated) != 1:
                raise allocerror, "allocation table corrupted, %s allocations and no free segments" % len(allocated)
            elif allocated[0][0] != 0: # sanity check
                raise allocerror, "allocation table corrupted, only allocation is %s" % _fmtr(*allocated[0])
            a,s = allocated[0]
            if addr != a+s: # sanity check
                raise allocerror, "invalid allocation %s, does not append to %s" % (_fmtr(addr,size), _fmtr(a,s))
            allocated[0] = (a,s+size)
            return
        
        endfree = not allocated or free[-1][0] > allocated[-1][0]
        if endfree:
            endaddr = free[-1][0]+free[-1][1]
        else:
            endaddr = allocated[-1][0]+allocated[-1][1]
            
        if addr > endaddr: # sanity check
            raise allocerror, "invalid allocation %s, last address is at 0x%X" % (_fmtr(addr,size), endaddr)
                
        if addr == endaddr: # special case: append
            if endfree: # sanity check
                raise allocerror, "invalid allocation %s, last free = %s" % (_fmtr(addr,size), _fmtr(*free[-1]))
            # enlarge last segment
            a,s = allocated[-1] 
            allocated[-1] = (a,s+size)
            return

        # retrieve the segment of free space
        fi = self._freei(addr)
        if fi < 0: # sanity check
            raise allocerror, "invalid allocation %s, first free = %s" % (_fmtr(addr,size), _fmtr(*free[0]))
        fa,fs = free[fi]

        previ = self._alloci(addr)
        nexti = previ+1
        
        if addr != fa: # sanity check
            raise allocerror, "invalid allocation %s, free = %s" % (_fmtr(addr,size), _fmtr(fa,fs))

        if previ >= 0: # special case: there is a previous segment
            # merge with previous
            pa,ps = allocated[previ]
            if pa+ps != addr: # sanity check
                raise allocerror, "allocation table corrupted 0x%X does not follow %s" % (addr, _fmtr(pa,ps))
            allocated[previ] = (pa,ps+size)
            i = previ
        else:
            if addr != 0: # sanity check
                raise allocerror, "first allocation can't be 0x%X" % _fmtr(addr,size)
            allocated.insert(0, (addr,size))
            i = 0
            nexti = 1

        ## TODO: optimize (common case first)
        if size > fs: # special case: size exceeds free segment size
            if fi != len(free)-1: # sanity check
                raise allocerror, "invalid allocation %s, free = %s" % (_fmtr(addr,size), _fmtr(fa,fs))
            # enlarge last free segment and fill it
            del free[-1]
        elif size == fs: # special case: free segment is filled exactly
            if nexti == len(allocated): # special case: no following segment
                if fi != len(free)-1: # sanity check
                    raise allocerror, "allocation table corrupted, no segments after free %s" % _fmtr(fa,fs)
            else:
                # combine with following segment
                na,ns = allocated[nexti]
                if addr+size != na: # sanity check
                    raise allocerror, "alloccation table corrupted, segments %s and %s apart" % (_fmtr(fa,fs),_fmtr(na,ns))
                a,s = allocated[i]
                allocated[i] = (a, ns+s)
                del allocated[nexti]
            del free[fi]
        else: # free segment is party filled
            free[fi] = (fa+size,fs-size)

    def free(self, addr, size):
        """free(addr, size) -> None
        
        Mark an allocated segment as "free".
        """
        if size == 0:
            return
        if addr < 0:
            raise allocerror, "invalid deallocation address %i" % addr
        if size < 0:
            raise allocerror, "invalid deallocation size %i" % size
        free = self._free
        allocated = self._allocated

        if not allocated: # sanity check
            raise allocerror, "nothing to deallocate"

        if not self.is_allocated(addr,size):
            raise allocerror, "%s is not allocated" % (_fmtr(addr,size))

        ai = self._alloci(addr)
        if ai < 0: # sanity check
            raise allocerror, "invalid deallocation %s, first allocation = %s" % (_fmtr(addr,size), _fmtr(*allocated[0]))
        aa,as = allocated[ai]
        if addr+size > aa+as: # sanity check
            raise allocerror, "invalid deallocation %s, alloc = %s" % (_fmtr(addr,size), _fmtr(aa,as))

        previ = self._freei(addr)
        nexti = previ+1


        if addr == aa: # special case: from beginning of allocation
            if previ >= 0: # there is a previous free segment
                pa,ps = free[previ]
                free[previ] = (pa,ps+size)
                i = previ
            else:
                free.insert(0, (addr,size))
                i = 0
                nexti = 1
            if size == as: # special case: free the whole segment
                if nexti < len(free):
                    a,s = free[i]
                    na,ns = free[nexti]
                    free[i] = (a,s+ns)
                    del free[nexti]
                del allocated[ai]
            else:
                allocated[ai] = (aa+size,as-size)
        else: # from middle of allocation
            allocated[ai] = (aa,addr-aa)            
            if addr+size == aa+as: # special case: free till the end of segment
                if nexti < len(free): # if there is a next free segment
                    na,ns = free[nexti]
                    free[nexti] = (addr, size+ns)
            else:
                free.insert(nexti, (addr,size))
                allocated.insert(ai+1, (addr+size,aa+as-addr-size))

    def check_integrity(self):
        """check_integrity() -> boolean

        Check that the allocation table hasn't been corrupted.
        (For debugging purposes)
        """
        all = self._free + self._allocated
        if not all:
            return 1
        free = self._free
        allocated = self._allocated
        all.sort()
        if all[0][0] != 0:
            return 0
        last = all[0]
        for x in all[1:]:
            if x[0] != last[0]+last[1]:
                return 0
            if (x in free and last in free) or (x in allocated and last in allocated):
                return 0
            last = x
        return 1

    def find_free(self, size):
        """find_free(size) -> (addr,size)

        Find a free segment of at least 'size' and return its
        address. This method does not mark the segment as "used",
        you have to do that separately with the allocate method.
        """
        free = self._free
        allocated = self._allocated
        if not free:
            if not allocated:
                return (0,size)
            return (allocated[-1][0]+allocated[-1][1],size)

        ## TODO: keep a list of free segments sorted by size and bisect instead of loop
        for a,s in free:
            if s >= size:
                return a,s

        endfree = not allocated or free[-1][0] > allocated[-1][0]
        if endfree:
            return (free[-1][0],size)
        else:
            endaddr = allocated[-1][0]+allocated[-1][1]
            return (endaddr,size)

    def alloc_greedy(self, size, maxsize=1.3):
        """alloc_greedy(size, maxsize=1.3) -> (addr,size)

        Similar to alloc_free, this method finds a suitable location
        and marks at least 'size' bytes as "used". However, if the
        total size of the continuous free segment is less than
        size * maxsize, the whole segment is allocated to prevent
        defragmentation.
        """
        a,s = self.find_free(size)
        if float(s)/size < maxsize:
            self.allocate(a,s)
            return (a,s)
        self.allocate(a,size)
        return (a,size)

    def alloc_free(self, size):
        """alloc_free(size) -> addr

        Marks exactly 'size' values as "used" and returns the address
        of the first value.
        """
        free = self.find_free(size)
        self.allocate(free[0],size)
        return free[0]

    def total_allocated(self):
        """total_allocated() -> int

        The total number of values marked as "used".
        """
        return reduce(lambda x,y: x+y[1], self._allocated, 0)
        
    def total_free(self):
        """total_free() -> int

        The total number of values marked as "free".
        """
        return reduce(lambda x,y: x+y[1], self._free, 0)

    def fragments(self):
        """fragments() -> int

        Return a number of continuous used/free areas
        in the allocation table.
        """
        return len(self._allocated) + len(self._free)

    def is_allocated(self, addr, size):
        """is_allocated(addr, size) -> boolean

        Check if all values in the given range are marked as "used".
        """
        allocated = self._allocated
        if not allocated:
            return 0
        i = self._alloci(addr)
        if i < 0:
            return 0
        a,s = allocated[i]
        if addr+size <= a+s:
            return 1
        return 0

    def is_free(self, addr, size):
        """is_free(addr, size) -> boolean

        Check if all values in the given range are marked as "free".
        """
        free = self._free
        if not free:
            return 0
        i = self._freei(addr)
        if i < 0:
            return 0
        a,s = free[i]
        if addr+size <= a+s:
            return 1
        return 0

    def __str__(self):
        free = self._free
        allocated = self._allocated
        l = free + allocated
        l.sort()
        result = []
        app = result.append
        for x in l:
            a,s = x
            if x in free:
                app("Free  ")
            else:
                app("Alloc ")
            app(_fmtr(a,s))
            app("  (%i bytes)" % s)
            app("\n")
        return "".join(result)

    def _freei(self, addr):
        return bisect(self._free, (addr, _MAXSIZE))-1

    def _alloci(self, addr):
        return bisect(self._allocated, (addr, _MAXSIZE))-1

    def _allocate(self, addr, size):
        old = self._free[:], self._allocated[:]
        self.allocate_(addr,size)
        if not self.check_integrity():
            print "ALLOCATION TABLE CORRUPTED"
            print "allocated %s" % _fmtr(addr,size)
            print "\nCurrent:"
            print str(self)
            self._free, self._allocated = old
            print "Old:"
            print str(self)
            raise allocerror, "integrity compromised"

    def _free(self, addr, size):
        old = self._free[:], self._allocated[:]
        self.free_(addr,size)
        if not self.check_integrity():
            print "ALLOCATION TABLE CORRUPTED"
            print "freed %s" % _fmtr(addr,size)
            print "\nCurrent:"
            print str(self)
            self._free, self._allocated = old
            print "Old:"
            print str(self)
            raise allocerror, "integrity compromised"
    

## ----- Random Access File ----- ##

class RAF(object):
    """Random Access File

    A utility class for reading and writing binary data
    to a file in random access mode.
    """
    
    def __init__(self, filename):
        """RAF(filename) -> RAF object

        Open a file in random access mode. If the file doesn't
        exist it is created.
        """
        fn = self.filename = filename
        f = file(fn, "ab")
        f.close()
        self.len = _filesize(self.filename)
        self.file = None

    def flush(self):
        """flush() -> None

        Flush the contents of the write buffer. Relies on the
        underlying file.flush implementation.
        """
        if self.file is None:
            return
        self.file.flush()

    def close(self):
        """close() -> None

        Close the file. Additional operations to the RAF
        object will re-open the file.
        """
        if self.file is None:
            return
        self.file.close()
        self.file = None

    def put(self, addr, data, size=None):
        """put(addr, data) -> None

        Write upto 'size' bytes from a string into the RAF at position
        'addr'. The size of the file is increased as needed. If 'data' is
        a file-like object (has a method called "read"), up to 'size' bytes
        are read from it and written to the RAF.
        """
        f = self._file()
        f.seek(addr)
        if not data:
            return
        if hasattr(data, "read"):
            if size is not None:
                left = size
            else:
                left = _BLOCKSIZE
            while 1:
                s = data.read(min(_BLOCKSIZE, left))
                if not s:
                    return
                if size is not None:
                    left -= len(s)
                if left <= 0:
                    return
                f.write(s)
        if size is None:
            f.write(data)
        else:
            f.write(data[:size])

    def get(self, addr, size):
        """get(addr, size) -> string

        Read upto 'size' bytes from the RAF at position 'addr'.
        """
        f = self._file()
        f.seek(addr)
        if size:
            return f.read(size)
        return ""

    def append(self, data):
        """append(data) -> None

        Append a string to the end of the RAF
        """
        f = self._file()
        f.seek(0,2)
        f.write(data)

    def fill(self, addr, repeats, data):
        """fill(addr, repeats, data) -> None

        Fill a segment starting from 'addr' by repeating
        the contents of data.
        """
        if not repeats:
            return
        mul = max(int(float(_BLOCKSIZE) / len(data)+0.5),1)
        mul = min(mul, repeats)
        f = self._file()
        f.seek(addr)
        left = repeats
        block = data * mul
        while 1:
            f.write(block)
            left -= mul
            if not left:
                return
            if left < mul:
                f.write(data * left)
                return

    def copy(self, fromaddr, toaddr, size):
        """copy(fromaddr, toaddr, size) -> None

        Copy data from one position to another.
        """
        if size <= _BLOCKSIZE:
            self.put(toaddr, self.get(fromaddr, size))
            return
        written = 0
        left = size
        while 1:
            tmp = self.get(fromaddr + written, min(left, _BLOCKSIZE))
            self.put(toaddr + written, tmp)
            written += len(tmp)
            left -= len(tmp)
            if not left:
                return

    def swap(self, fromaddr, toaddr, size):
        """swap(fromaddr, toaddr, size) -> None

        Swap the data stored in two position.
        """
        if size <= _BLOCKSIZE:
            tmp = self.get(fromaddr, size)
            self.copy(toaddr, fromaddr, size)
            self.put(toaddr, tmp)
            return
        written = 0
        left = size
        while 1:
            bsize = min(left, _BLOCKSIZE)
            tmp = self.get(fromaddr+written, bsize)
            self.copy(toaddr+written, fromaddr+written, len(tmp))
            self.put(toaddr, tmp)
            written += len(tmp)
            left -= len(tmp)
            if not left:
                return

    def _open(self):
        self.file = file(self.filename, "r+b")

    def _file(self):
        if self.file is None:
            self._open()
        return self.file

    def _checksize(self):
        f = self._file()
        f.seek(0,2)
        self.len = f.tell()
        return self.len


class FileHeader(object):
    """Generic header object for storing indexing etc. information"""



## ----- File Memory ----- ##

class FileMem(object):
    """File Memory

    A random access file from which you can allocate chunks
    using C-like malloc, realloc and free.
    """
    
    _ptr = "!2Q"
    _allocblock = 16
    pickle_proto = -1
    
    def __init__(self, filename, allocblock = None):
        if allocblock is not None:
            self._allocblock = allocblock
        self.filename = filename        
        self.header = None
        if not exists(filename):
            self._new()
        else:
            self.file = RAF(self.filename)
            self._load_header()

    def flush(self):
        """flush() -> None

        Flush the contents of the write buffer. Relies on the
        underlying file.flush implementation.
        """
        self._save_header()
        self.file.flush()

    def close(self):
        """close() -> None

        Closes the file.
        """
        self._save_header()
        self.file.close()

    def alloc(self, size):
        """alloc(size) -> (address, size)

        Allocates an area of the file that has at least 'size' bytes.
        The actual allocated size will be returned and should be kept
        for deallocation purposes.
        """
        block = self._allocblock
        realsize = int(math.ceil(size / float(block)))*block
        return self.header.alloctab.alloc_greedy(realsize)

    def free(self, addr, size):
        """free(address, size) -> None

        Free a previously allocated area of the file.
        """
        self.header.alloctab.free(addr,size)

    def realloc(self, addr, size, new_size):
        """realloc(address, size, newsize) -> (address, size)

        Copy the contents of an address area to a new address and
        reserve more bytes after the data.
        """
        self.free(addr,size)
        a,s = self.alloc(new_size)
        if a != addr:
            self.file.copy(addr, a, size)
        return a,s

    def _new(self):
        self.file = RAF(self.filename)
        self.header = FileHeader()
        a = self.header.alloctab = AllocTab()
        sz = struct.calcsize(self._ptr)
        a.allocate(0, sz)
        self._hdrptr = None

    def _get_hdrptr(self):
        ptrs = self.file.get(0, struct.calcsize(self._ptr))
        a,s = struct.unpack(self._ptr, ptrs)
        self._hdrptr = a,s
        return a,s

    def _put_hdrptr(self):
        self.file.put(0, struct.pack(self._ptr, *self._hdrptr))

    def _load_header(self):
        a,s = self._get_hdrptr()
        f = self.file
        f.get(a,0)
        self.header = cPickle.load(f._file())

    def _save_header(self):
        f = self.file
        tabs = cPickle.dumps(self.header, self.pickle_proto)
        if self._hdrptr:
            a,s = self._hdrptr
            if len(tabs) > s:                
                self.free(a,s)
                s = int(len(tabs)*1.5)
                a,s = self.alloc(s)
        else:
            s = int(len(tabs)*1.5)
            a,s = self.alloc(s)
        tabs = cPickle.dumps(self.header, self.pickle_proto)
        assert len(tabs) <= s
        f.put(a,tabs)
        self._hdrptr = (a,s)
        self._put_hdrptr()



## ----- ID Table ----- ##

class IDTable(object):
    """ID Table

    Tracks ID numbers in a scalable way (O(1) access via id)
    """
    _ptrtype = FileMem._ptr
    _ptrsz = struct.calcsize(_ptrtype)
    _initsize = 100
    
    def __init__(self, raf):
        self._alloc = AllocTab()
        self._raf = raf
        self._new()

    def _new(self):
        cap = self._capacity = self._initsize
        sz = struct.calcsize(self._ptrtype) * cap
        self._ptr = self._raf.alloc(sz)

    def __getstate__(self):
        state = {}
        state.update(self.__dict__)
        del state['_raf']
        return state

    def _grow(self):
        cap = self._capacity
        sz = self._ptrsz * cap
        new_cap = int(cap * 1.5)
        new_sz = self._ptrsz * new_cap
        sz *= cap
        addr,size = self._ptr
        self._ptr = self._raf.realloc(addr, size, new_sz)
        self._capacity = new_cap

    def new_id(self):
        a = self._alloc.alloc_free(1)
        if a >= self._capacity:
            self._grow()
        self[a] = (0,0)
        return a

    def __getitem__(self, id):
        if not isinstance(id, int):
            raise TypeError, "id must be an integer"
        sz = self._ptrsz
        if id >= self._capacity or id < 0 or id not in self:
            raise ValueError, "invalid id: %s" % id
        addr = self._ptr[0] + id * sz
        data = self._raf.file.get(addr, sz)
        ptr = struct.unpack(self._ptrtype, data)
        return ptr

    def __setitem__(self, id, ptr):
        if not isinstance(id, int):
            raise TypeError, "id must be an integer"
        if id >= self._capacity or id < 0 or id not in self:
            raise ValueError, "invalid id: %s" % id
        data = struct.pack(self._ptrtype, *ptr)
        addr = self._ptr[0] + id * self._ptrsz
        self._raf.file.put(addr, data)

    def __contains__(self, id):
        if not isinstance(id, int):
            raise TypeError, "id must be an integer"
        return self._alloc.is_allocated(id, 1)

    def __delitem__(self, id):
        if not isinstance(id, int):
            raise TypeError, "id must be an integer"
        if id not in self:
            raise ValueError, "invalid id %s" % id
        self[id] = (0,0)
        self._alloc.free(id, 1)

    def __len__(self):
        return self._alloc.total_allocated()

    def iterids(self):
        for a,s in self._alloc._allocated:
            for x in xrange(s):
                yield a+x
        

## ----- Object Storage ----- ##

class ObjectStorage(FileMem):
    """Object Storage

    File storage for saving Python objects (using pickle) and referencing
    them by an id number. The Object Storage can be used like a dictionary,
    except that new ids must be reserved using new or new_id before assignment.

    The ObjectStorage is still pretty low level, and is meant to be used as
    the base implementation for more advanced persistence systems.

    Example:
    
    >>> storage = ObjectStorage("mystorage")
    >>> obj = [1,2,3]
    >>> id = storage.new(obj)
    >>> print storage[id]
    [1,2,3]
    >>> del storage[id]
    >>> print id in storage
    False

    or using new_id()...

    >>> storage = ObjectStorage("mystorage")
    >>> obj = [1,2,3]
    >>> id = storage.new_id()
    >>> storage[id] = obj
    """
    pickle_proto = -1

    def has_id(self, id):
        return id in self.header.idtab

    def iterids(self):
        return self.header.idtab.iterids()

    def new_id(self):
        return self.header.idtab.new_id()

    def new(self, obj):
        """new(obj) -> id
        
        Puts an object into the db and creates a new id for it"""
        data = cPickle.dumps(obj, self.pickle_proto)
        idtab = self.header.idtab
        id = idtab.new_id()
        s = len(data)
        a,s = self.alloc(s)
        self.file.put(a,data)
        idtab[id] = (a,s)
        return id

    def get_pickle(self, id):
        idtab = self.header.idtab
        a,s = idtab[id]
        if (a,s) == (0,0):
            raise ValueError, "unallocated id %i" % id
        data = self.file.get(a,s)
        return data

    def set_pickle(self, id, data):
        idtab = self.header.idtab
        a,s = idtab[id]
        if len(data) > s:
            new_s = int(len(data)*1.5)
            self.free(a,s)
            a,s = self.alloc(new_s)
            idtab[id] = (a,s)
        elif len(data) < s*0.2: # if the pickle is considerably smaller, save space
            new_s = int(len(data)*1.5)
            self.free(a,s)
            a,s = self.alloc(new_s)
            idtab[id] = (a,s)
        self.file.put(a, data)

    def __setitem__(self, id, obj):
        data = cPickle.dumps(obj, self.pickle_proto)
        self.set_pickle(id, data)

    def __getitem__(self, id):
        data = self.get_pickle(id)
        return cPickle.loads(data)

    def __delitem__(self, id):
        idtab = self.header.idtab
        a,s = idtab[id]
        self.free(a,s)
        del idtab[id]

    def __contains__(self, id):
        return self.has_id(id)

    def __len__(self):
        """number of objects"""
        return len(self.header.idtab)

    
    def _new(self):
        FileMem._new(self)
        self.header.idtab = IDTable(self)

    def _load_header(self):
        FileMem._load_header(self)
        self.header.idtab._raf = self



class _Tier(dict):
    def __init__(self):
        self._next = None  ## id of next dictionary or None
        self._last_access = {}

    def __setitem__(self, key, value):
        dict.__setitem__(key, value)
        self._last_access[key] = time.time()

    def __getitem__(self, key):
        value = dict.__getitem__(self, key)
        self._last_access[key] = time.time()
        return value

    def __contains__(self, key):
        c = dict.__contains__(self, key)
        if c:
            self._last_access[key] = time.time()

    has_key = __contains__

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        del self._last_access[key]

    def update(self, src):
        for k,v in src:
            self[k] = v

    def clear(self):
        dict.clear(self)
        self._last_access.clear()


## ----- Tiered Dictionary ----- ##

class TierDict(object):
    """Tiered Dictionary (WORK IN PROGRESS)

    Persistent dictionary that is divided into constant sized "tiers" and
    only maximum of two tiers are kept in memory at once. If keys are not
    accessed, they gradually drop into lower tiers.
    """
    
    def __init__(self, objstor, limit=1000):
        self._storage = objstor
        self._first = _Tier()
        self._limit = limit

    def __setitem__(self, key, value):
        self._first[key] = value
        ## TODO: check first size

    def __getitem__(self, key):
        tier = self._first
        while key not in tier:
            if tier._next is None:
                raise KeyError, key
            id = tier._next
            tier = self._storage[id]
        value = tier[key]
        if tier is not self._first: ## move to first tier
            del tier[key]
            self._storage[id] = tier
            self._first[key] = value
            ## check size of first
        return value
                

    def __delitem__(self, key):
        id = None
        tier = self._first
        while 1:
            if key in tier:
                del tier[key]
                if id is not None:
                    self._storage[id] = tier
            id = tier._next
            if id is None:
                return
            tier = self._storage[id]
            

    def update(self, src):
        for key in src:
            self[key] = src[key]

    def clear(self):
        id = self._first._next
        self._first.clear()
        while id is not None:
            next = self._storage[id]._next
            del self._storage[id]
            id = next

    def __contains__(self, key):
        tier = self._first
        while key not in tier:
            id = tier._next
            if id is None:
                return False
            tier = self._storage[id]
        if tier is not self._first: ## move to first tier
            del tier[key]
            self._storage[id] = tier
            self._first[key] = value
            ## check size of first
        return True

    has_key = __contains__

    def iterkeys(self):
        pass

    def itervalues(self):
        pass

    def iteritems(self):
        pass

    def keys(self):
        pass

    def items(self):
        pass

    def values(self):
        pass


class Shelve(object):
    """Shelve

    A pure Python shelve implementation that uses the Object Storage.
    Still has the same limitation as dumbdb, that all keys are loaded
    into memory when the shelve is opened, but otherwise it works much
    faster and is more reliable.
    """
   
    def __init__(self, filename):
        self._storage = ObjectStorage(filename)
        if len(self._storage) == 0:
            self._new()
        else:
            self._load_keys()

    def _new(self):
        self._keys = {}
        id = self._storage.new(self._keys)
        assert id == 0
        self._storage.flush()

    def _load_keys(self):
        self._keys = self._storage[0]
        
    def _save_keys(self):
        self._storage[0] = self._keys

    def flush(self):
        self._save_keys()
        self._storage.flush()

    def close(self):
        self.flush()
        self._storage.close()
        self._storage = None

    def open(self, filename):
        if self._storage is not None:
            self.close()
        self.__init__(filename)

    def __setitem__(self, key, value):
        keys = self._keys
        if key not in keys:
            id = self._storage.new(value)
            keys[key] = id
            return
        id = keys[key]
        self._storage[id] = value

    def __getitem__(self, key):
        id = self._keys[key]
        return self._storage[id]

    def __delitem__(self, key):
        id = self._keys[key]
        del self._storage[id]

    def __iter__(self):
        return self.iterkeys()

    def __contains__(self, key):
        return key in self._keys

    def has_key(self, key):
        return self._keys.has_key(key)

    def keys(self):
        return self._keys.keys()

    def values(self):
        return [self._storage[id] for id in self._keys.values()]

    def items(self):
        return zip(self.keys(), self.values())

    def iterkeys(self):
        return self._keys.iterkeys()

    def itervalues(self):
        for id in self._keys.itervalues():
            yield self._storage[id]

    def iteritems(self):
        for key,id in self._keys.iteritems():
            yield (key, self._storage[id])




# -*- coding: utf-8 -*-

"""
$Id: module_calc.py 331 2012-11-13 10:18:18Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/module_calc.py $
"""

import re
from operator import pow, add, sub, mul, div, mod

import urllib2
import requests
import json


def doTheMath(matchobj):
    e = matchobj.groups()
    res = {"^": pow, "+": add, "-": sub, "*": mul, "/": div, "%": mod}[e[1]](float(e[0]), float(e[2]))
    return str(res)


def calc(str, match=True):
    if match:
        str = str.group(0)[1:-1]

    while True:
        newstr = re.sub("[(][-+*^%/\d.]*[)]", calc, str)
        if (newstr == str):
            break
        else:
            str = newstr

    for op in ["[\^]", "[*/%]", "[-+]"]:
        while True:
            newstr = re.sub("(\-?[\d.]+)(%s)(\-?[\d.]+)" % op, doTheMath, str)
            if (newstr == str):
                break
            else:
                str = newstr

    return str


def calc_google(args):
    google_url = "http://www.google.com/search?hl=en&num=1&q=%s"
    search_url = google_url % urllib2.quote(args)

    request = urllib2.Request(search_url)
    request.add_header('User-Agent', 'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.9.1.4) Gecko/20091016 Firefox/3.5.4')

    opener = urllib2.build_opener()

    d = opener.open(request).read()

    m = re.search("calculator-40.gif[^>]+><td>[^<]+<td[^>]+><h2[^>]+><b>([^<]+)</b>", d)
    if m:
        res = m.group(1)
        # clean up
        res = res.replace("<font size=-2> </font>", " ")
        res = res.replace("<sup>", "^").replace("</sup>", "")
        res = res.replace("&#215;", "×")
        return res
    else:
        return "Invalid calculation"


def calc_google_ig(args):
    google_url = "http://www.google.com/ig/calculator?hl=en&q=%s"
    search_url = google_url % urllib2.quote(args)
    headers = {'User-Agent': 'Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.9.1.4) Gecko/20091016 Firefox/3.5.4'}
    r = requests.get(search_url, headers=headers)
    d = r.text
    d = d.replace('lhs', '"lhs"', 1).replace('rhs', '"rhs"', 1).replace('error', '"error"', 1).replace('icc', '"icc"', 1)
    res = json.loads(d)
    if not res['error']:
        result = "%s = %s" % (res['lhs'], res['rhs'])
        return result.encode('ascii', 'ignore')
    else:
        return "Invalid calculation"


def command_calc(bot, user, channel, args):
    if not args:
        return
    return bot.say(channel, calc_google_ig(args))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""pyfibot module_rss.py
@author Henri 'fgeek' Salo <henri@nerv.fi>, Tomi 'dmc' Nykänen
@copyright Copyright (c) 2010 Henri Salo
@licence BSD

$Id: module_rss.py 332 2012-11-13 10:38:14Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/module_rss.py $
"""

import sys
import os
import re
import urllib
import logging
import logging.handlers
import hashlib
from threading import Thread
import urllib2
import sqlite3

# import py2.6 json if available, fall back to simplejson
try:
    import json
except:
    try:
        import simplejson as json
    except ImportError, error:
        print('Error starting rss module: %s' % error)
        init_ok = False

try:
    import feedparser
    from twisted.internet import reactor
    import yaml
    import htmlentitydefs
    init_ok = True
except ImportError, error:
    print('Error starting rss module: %s' % error)
    init_ok = False

# Initialize logger
log = logging.getLogger('rss')

t = None
t2 = None
indexfeeds_callLater = None
output_callLater = None

# Config format:
# database: rss.db
# delays:
#   rss_sync: 300 #How often we synchronize rss-feeds (in seconds)
#   output: 7 #How often to output new elements to channels
# output_syntax: 0 #0 == feed_title: title - url, 1 == feed_title: title - shorturl, 2 == feed_title: title (id), 3 == feed_title: title, 4 == title
# bitly_login: #Needed if using shorturl format
# bitly_api_key: #Needed if using shorturl format

# With output_syntax #2 you could get url via .url <id>


def event_signedon(bot):
    """Starts rotators"""
    global indexfeeds_callLater, output_callLater
    if not init_ok:
        log.error("Config not ok, not starting rotators")
        return False
    if (empty_database > 0):
        if (indexfeeds_callLater != None):
            log.info("Stopping previous indexfeeds thread")
            indexfeeds_callLater.cancel()
        rotator_indexfeeds(bot, rssconfig["delays"]["rss_sync"])
        if (output_callLater != None):
            log.info("Stopping previous output thread")
            output_callLater.cancel()
        rotator_output(bot, rssconfig["delays"]["output"])


def init(botconfig):
    """Creates database if it doesn't exist"""
    if not init_ok:
        log.error("Config not ok, skipping init")
        return False

    global rssconfig
    # Read configuration
    configfile = os.path.join(sys.path[0], 'modules', 'module_rss.conf')
    rssconfig = yaml.load(file(configfile))

    db_conn = sqlite3.connect(rssconfig["database"])
    d = db_conn.cursor()
    d.execute("CREATE TABLE IF NOT EXISTS feeds (id INTEGER PRIMARY KEY,feed_url TEXT,channel TEXT,feed_title TEXT, output_syntax_id INTEGER);")
    d.execute("CREATE TABLE IF NOT EXISTS titles_with_urls (id INTEGER PRIMARY KEY,feed_url TEXT,title TEXT,url TEXT,channel TEXT,printed INTEGER,hash TEXT UNIQUE);")
    db_conn.commit()
    #Check if database is empty
    global empty_database
    empty_database = d.execute("SELECT COUNT(*) FROM feeds").fetchone()[0]
    d.close()


def rss_addfeed(bot, user, channel, feed_url, output_syntax):
    """Adds RSS-feed to sqlite-database"""
    global empty_database
    try:
        feed_data = feedparser.parse(feed_url)
        feed_title = feed_data['feed']['title']
    except KeyError, e:
        return bot.say(channel, "Nothing inserted to database. Probably you mistyped URL?")
    """Initialize connection to database-file"""
    db_conn = sqlite3.connect(rssconfig["database"])
    d = db_conn.cursor()

    """Lets create scheme if the database-file does not exist"""
    try:
        fileinfo = os.stat(rssconfig["database"])

        if not os.path.isfile(rssconfig["database"]) or fileinfo.st_size == 0:
            # TODO: Update schema creation after database schema is ready
            d.execute("CREATE TABLE feeds int primary key unique, url text)")
            db_conn.commit()
            bot.say(channel, "Database \"%s\" created." % rssconfig["database"])
    except Exception, e:
        bot.say(channel, "Error: %s" % e)

    d.execute("SELECT * FROM feeds WHERE feed_url = ? AND channel = ?", (feed_url, channel, ))
    already_on_db = d.fetchone()
    if already_on_db is None:
        data = [None, feed_url, channel, feed_title, output_syntax]
        d.execute("INSERT INTO feeds VALUES (?, ?, ?, ?, ?)", data)
        db_conn.commit()
        d.close()
        if (empty_database == 0):
            rotator_indexfeeds(bot, rssconfig["delays"]["rss_sync"])
            rotator_output(bot, rssconfig["delays"]["output"])
            empty_database = 1
            return bot.say(channel, "Url \"%s\" inserted to database." % feed_url)
    else:
        id = already_on_db[0]
        return bot.say(channel, "Url \"%s\" is already on database with ID: %i" % (feed_url, id))


def rss_delfeed(bot, user, channel, args):
    """Deletes RSS-feed from sqlite-database by given ID. Should have support for deletion using ID or feed's URL"""
    db_conn = sqlite3.connect(rssconfig["database"])
    d = db_conn.cursor()

    d.execute("SELECT id, feed_url, channel FROM feeds WHERE id = ? OR feed_url = ?", (args, args,))
    row = d.fetchone()
    feed_url = row[1]
    if (db_conn.execute("DELETE FROM feeds WHERE channel = ? AND id = ? OR channel = ? AND feed_url = ?", (channel, args, channel, args)).rowcount == 1):
        db_conn.commit()
        bot.say(channel, "Feed %s deleted successfully" % feed_url.encode("UTF-8"))
    d.close()


def list_feeds(channel):
    """Lists channels (or all) RSS-feeds from sqlite-database. Initialize connection to database-file"""
    db_conn = sqlite3.connect(rssconfig["database"])
    d = db_conn.cursor()

    feeds = []
    if (channel == -1):
        d.execute("SELECT id, feed_url, channel FROM feeds ORDER BY id")
    else:
        d.execute("SELECT id, feed_url, channel FROM feeds WHERE channel = ? ORDER BY id", (channel,))
    for row in d:
        id = row[0]
        feed_url = row[1]
        feed_url = feed_url.encode('UTF-8')
        channel = row[2]
        data = [id, feed_url, channel]
        feeds.append(data)
    d.close()
    return feeds


def rss_modify_feed_setting(bot, feed_ident, channel, target, tvalue):
    """Modifies feed's settings"""
    db_conn = sqlite3.connect(rssconfig["database"])
    d = db_conn.cursor()
    d.execute("SELECT id, feed_url FROM feeds WHERE id = ? OR feed_url = ?", (feed_ident, feed_ident,))
    row = d.fetchone()
    feed_url = row[1].encode("UTF-8")
    if (target == "title"):
        if (db_conn.execute("UPDATE feeds SET feed_title = ? WHERE id = ? OR feed_url = ? AND channel = ?", (tvalue, feed_ident, feed_ident, channel)).rowcount == 1):
            db_conn.commit()
            bot.say(channel, "Feed's (%s) title modified to \"%s\"" % (feed_url, tvalue.encode("UTF-8")))
    elif (target == "syntax"):
        if (db_conn.execute("UPDATE feeds SET output_syntax_id = ? WHERE id = ? OR feed_url = ? AND channel = ?", (tvalue, feed_ident, feed_ident, channel)).rowcount == 1):
            db_conn.commit()
            bot.say(channel, "Feed's (%s) output syntax modified to \"%s\"" % (feed_url, tvalue.encode("UTF-8")))
    else:
        bot.say(channel, "Invalid syntax! Usage: Add feed: .rss add <feed_url>, Delete feed: .rss del <feed_url/feed_id>, List feeds: .rss list, Change feed settings: .rss set <feed_id/feed_url> title/syntax <value>")
        d.close()


def rss_listfeeds(bot, user, channel, args):
    """Lists all RSS-feeds added to channel"""
    feeds = list_feeds(channel)
    for feed in feeds:
        bot.say(channel, "%s: %s" % (feed[0], feed[1]))


def command_rss(bot, user, channel, args):
    """Usage: Add feed: .rss add <feed_url>, Delete feed: .rss del <feed_url/feed_id>, List feeds: .rss list, Change feed settings: .rss set <feed_id/feed_url> title/syntax <value>"""
    try:
        args = args.split()
        subcommand = args[0]
        if (subcommand != "list"):
            feed_ident = args[1]
        if (isAdmin(user)):
            if (subcommand == "add"):
                if (len(args) > 2):
                    output_syntax = args[2]
                else:
                    output_syntax = rssconfig["output_syntax"]
                rss_addfeed(bot, user, channel, feed_ident, output_syntax)
            elif (subcommand == "del"):
                rss_delfeed(bot, user, channel, feed_ident)
            elif (subcommand == "list"):
                rss_listfeeds(bot, user, channel, None)
            elif (subcommand == "set"):
                target = args[2]
                tvalue = args[3].decode("UTF-8")
                if (target == "title"):
                    rss_modify_feed_setting(bot, feed_ident, channel, "title", tvalue)
                elif (target == "syntax"):
                    rss_modify_feed_setting(bot, feed_ident, channel, "syntax", tvalue)
            else:
                bot.say(channel, "Invalid syntax! Usage: Add feed: .rss add <feed_url>, Delete feed: .rss del <feed_url/feed_id>, List feeds: .rss list, Change feed settings: .rss set <feed_id/feed_url> title/syntax <value>")
    except IndexError:
        bot.say(channel, "Invalid syntax! Usage: Add feed: .rss add <feed_url>, Delete feed: .rss del <feed_url/feed_id>, List feeds: .rss list, Change feed settings: .rss set <feed_id/feed_url> title/syntax <value>")


def shorturl(url):
    try:
        req = urllib2.Request("http://api.bit.ly/v3/shorten?%s" % urllib.urlencode({'longUrl': url, 'login': rssconfig["bitly_login"], 'apiKey': rssconfig["bitly_api_key"], 'format': 'json'}))
        results = json.loads(urllib2.urlopen(req).read())
        #log.debug("Shorturl: %s" % results['id'].encode("UTF-8"))
        if (results['status_code'] == 200):
            return results['data']['url'].encode("UTF-8")
        raise Exception("Error in function shorturl: %s" % results['status_txt'])
    except HTTPError, e:
        log.debug('Error in function shorturl (url => %s): %s' % (url, e.read()))


def unescape(text):
    """Unescape ugly wtf-8-hex-escaped chars."""
    def fixup(m):
        text = m.group(0)
        if text[:2] == "&#":
            # character reference
            try:
                if text[:3] == "&#x":
                    return unichr(int(text[3:-1], 16))
                else:
                    return unichr(int(text[2:-1]))
            except ValueError:
                pass
        else:
            # named entity
            try:
                text = unichr(htmlentitydefs.name2codepoint[text[1:-1]])
            except KeyError:
                pass
        return text  # leave as is
    return re.sub("&#?\w+;", fixup, text)


def remove_html_tags(data):
    p = re.compile(r'<.*?>')
    return p.sub('', data)


def sqlite_add_item(bot, feed_url, title, url, channel, cleanup):
    """Adds item with feed-url, title, url, channel and marks it as non-printed"""
    if not feed_url and title and url:
        log.debug('Arguments missing in sqlite_add_item.')
        return
    try:
        db_conn = sqlite3.connect(rssconfig["database"])
        d = db_conn.cursor()
        hash_string = url + "?channel=" + channel
        data = [None, feed_url, title, url, channel, cleanup, hashlib.md5(hash_string).hexdigest()]
        d.execute("INSERT INTO titles_with_urls VALUES (?, ?, ?, ?, ?, ?, ?)", data)
        log.info('Added title \"%s\" with URL \"%s\" (%s) to channel %s to database.' % (title, url, feed_url, channel))
        id = d.lastrowid
        db_conn.commit()
        db_conn.close()
    except sqlite3.IntegrityError, e:
        # Couldn't add entry twice
        return
    except Exception, e:
        # Database is already opened
        log.debug('Error in sqlite_add_item: %s' % e)
        pass


def indexfeeds(bot):
    """Updates all RSS-feeds found on database and outputs new elements"""
    try:
        cleanup = 0
        log.debug("indexfeeds thread started")
        db_conn = sqlite3.connect(rssconfig["database"])
        d = db_conn.cursor()

        # If table is removed, then create a new one
        titles_table_exists = d.execute("SELECT count(*) FROM sqlite_master WHERE type='table' AND name='titles_with_urls'").fetchone()
        if (titles_table_exists[0] == 0):
            init(None)

        # Let's count all rows to index
        rowcount = 0
        rows = d.execute("SELECT id, feed_url, channel FROM feeds ORDER BY id")
        for row in rows:
            rowcount = rowcount + 1
        log.debug('Feed count is: %i' % rowcount)
        feeds = list_feeds(-1)
        for feed in feeds:
            id = feed[0]
            feed_url = feed[1]
            log.debug('Indexing feed: %s' % feed_url)
            channel = feed[2]
            # If first run of current feed, insert new elements as "printed" so bot won't flood whole feed on startup/insert
            cleanup = 0
            titles_count = d.execute("SELECT count(*) FROM titles_with_urls WHERE feed_url = ?", (feed_url,)).fetchone()
            if (titles_count[0] == 0):
                cleanup = 1

            feed_data = feedparser.parse(feed_url)
            feed_data.entries.reverse()
            for entry in feed_data.entries:
                try:
                    title = remove_html_tags(entry['title'])
                    url = entry['link']
                    sqlite_add_item(bot, feed_url, title, url, channel, cleanup)
                except KeyError, e:
                    log.debug('indexfeeds: Keyerror %s' % e)
                except Exception, e:
                    log.debug('indexfeeds first: Exception %s' % e)
        db_conn.close()
        log.debug("indexfeeds thread terminated")
    except Exception, e:
        log.debug('Error in indexfeeds: %s', e)


def command_url(bot, user, channel, args):
    """Prints feed's element url by given id"""
    id = args
    try:
        db_conn = sqlite3.connect(rssconfig["database"])
        d = db_conn.cursor()
        d.execute("SELECT * FROM titles_with_urls WHERE id = ?", (id,))
        row = d.fetchone()
        if (row != None):
            id = row[0]
            feed_url = row[1]
            title = row[2]
            url = row[3]
            channel = row[4].encode("UTF-8")
            url = url.encode("UTF-8")
            bot.say(channel, "%s" % (url))
    except Exception, e:
        # Database is already opened
        log.debug("Exception in command showurl: %s" % e)


def output(bot):
    """This function is launched from rotator to collect and announce new items from feeds to channel"""
    try:
        db_conn = sqlite3.connect(rssconfig["database"])
        d = db_conn.cursor()
        d.execute("SELECT * FROM titles_with_urls WHERE printed='0'")
        row = d.fetchone()
        if (row != None):
            log.debug("New row found for output")
            log.debug(row)

            id = row[0]
            feed_url = row[1]
            feed_output_syntax = d.execute("SELECT output_syntax_id FROM feeds WHERE feed_url = ?", (feed_url,)).fetchone()[0]
            if (feed_output_syntax == None):
                feed_output_syntax = rssconfig["output_syntax"]

            title = row[2]
            url = row[3]
            channel = row[4]

            title = unicode(unescape(title)).encode("UTF-8")
            channel = channel.encode("UTF-8")
            url = url.encode("UTF-8")

            feed_title = d.execute("SELECT feed_title from feeds where feed_url = ?", (feed_url,)).fetchone()[0].encode('UTF-8')

            if (feed_output_syntax == 0):
                bot.say(channel, "%s: %s – %s" % (feed_title, title, url))
            elif (feed_output_syntax == 1):
                bot.say(channel, "%s: %s – %s" % (feed_title, title, shorturl(url)))
            elif (feed_output_syntax == 2):
                bot.say(channel, "%s: %s (%i)" % (feed_title, title, id))
            elif (feed_output_syntax == 3):
                bot.say(channel, "%s: %s" % (feed_title, title))
            elif (feed_output_syntax == 4):
                bot.say(channel, "%s" % title)

            data = [url, channel]
            d.execute("UPDATE titles_with_urls SET printed=1 WHERE URL=? and channel=?", data)
            db_conn.commit()
            log.debug("output thread terminated cleanly")
    except StopIteration:
        pass
    except Exception, e:
        # Database is already opened
        log.debug("Exception in function output: %s" % e)
        pass


def rotator_indexfeeds(bot, delay):
    """Timer for methods/functions"""
    try:
        global t, t2, indexfeeds_callLater
        if (type(t2).__name__ == 'NoneType'):
            t = Thread(target=indexfeeds, args=(bot,))
            t.daemon = True
            t.start()
        elif t2.isAlive() == False:
            t = Thread(target=indexfeeds, args=(bot,))
            t.daemon = True
            t.start()
        if (empty_database > 0):
            indexfeeds_callLater = reactor.callLater(delay, rotator_indexfeeds, bot, delay)
    except Exception, e:
        print e


def rotator_output(bot, delay):
    """Timer for methods/functions"""
    try:
        global t, t2, output_callLater
        if t.isAlive() == False:
            t2 = Thread(target=output, args=(bot,))
            t2.daemon = True
            t2.start()
            t2.join()
        if (empty_database > 0):
            output_callLater = reactor.callLater(delay, rotator_output, bot, delay)
    except Exception, e:
        print e



##
## ODB2 - Pure Python object database
## Copyright (C) 2004 Sami Hangaslammi (shang@jyu.fi)
##

import time

class TrackingDict(dict):
    def __init__(self, *args, **kw):
        dict.__init__(self, *args, **kw)
        self._last_access = {}

    def __getitem__(self, key):
        self._last_access[key] = time.time()
        return dict.__getitem__(self, key)

    def __setitem__(self, key, value):
        self._last_access[key] = time.time()
        return dict.__setitem__(self, key, value)

    def __delitem__(self, key):
        del self._last_access[key]
        return dict.__delitem__(self, key)

    def setdefault(self, key, value):
        self._last_access[key] = time.time()
        return dict.setdefault(self, key, value)

    def del_old(self, timelimit):
        limit = time.time() - timelimit
        deletions = [key for key,t in self._last_access.iteritems()
                     if t < limit]
        map(self.__delitem__, deletions)

    def clear(self):
        dict.clear(self)
        self._last_access.clear()



"""
$Id: module_oraakkeli.py 331 2012-11-13 10:18:18Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/available/module_oraakkeli.py $
"""

import urllib


def command_oraakkeli(bot, user, channel, args):
    """Asks a question from the oracle (http://www.lintukoto.net/viihde/oraakkeli/)"""
    if not args:
        return
    args = urllib.quote_plus(args)
    answer = getUrl("http://www.lintukoto.net/viihde/oraakkeli/index.php?kysymys=%s&html=0" % args).getContent()
    answer = unicode(answer)
    answer = answer.encode("utf-8")
    return bot.say(channel, "Oraakkeli vastaa: %s" % answer)



# -*- encoding: utf-8 -*-

"""
$Id: module_expl.py 331 2012-11-13 10:18:18Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/module_expl.py $
"""

import os
import os.path
import sys
import re
import random
import fnmatch


def expl_parseterm(expl):
    expl = expl.split(" ")
    expl = expl[0]
    expl = expl.lower()
    expl = expl.strip()
    invalidchars = re.compile("[^a-z0-9\ :\.-]")
    expl = invalidchars.sub("_", expl)
    return expl


def expl_getdir(channel):
    expldir = os.path.join(sys.path[0], "expl", channel)
    if not os.path.exists(expldir):
        return None
    return expldir


def expl_getlist(expldir):
    list = os.listdir(expldir)
    dotfile = re.compile(r'(^[^\.])')
    list = filter(dotfile.match, list)
    return list


def expl_getexpl(expldir, term):
    f = file(os.path.join(expldir, term))
    expl = f.read()
    f.close()
    return expl


def check_params(bot, args, channel):
    """Do some initial checking for the stuff we need for every subcommand"""
    if not args:
        return False

    expldir = expl_getdir(channel)
    if not expldir:
        bot.log("No expldir for channel %s, create %s to enable expl." % (channel, os.path.join(sys.path[0], "expl", channel)))
        return False

    termlist = expl_getlist(expldir)

    return expldir, termlist


def command_expl(bot, user, channel, args):
    """Explains terms. Usage: .expl <term> See also: .rexpl, .add, .del, .ls"""

    try:
        expldir, termlist = check_params(bot, args, channel)
    except TypeError:
        return

    term = expl_parseterm(args)
    if term in termlist:
        expl = expl_getexpl(expldir, term)
        return bot.say(channel, "'%s': %s" % (term, expl))
    else:
        return bot.say(channel, "Term '%s' not found. Try .help ls" % term)


def command_rexpl(bot, user, channel, args):
    """Returns random explanation. See also: .expl, .add, .del, .ls"""

    try:
        expldir, termlist = check_params(bot, "none", channel)
    except TypeError:
        return

    term = termlist[random.randrange(0, len(termlist) - 1)]
    expl = expl_getexpl(expldir, term)
    return bot.say(channel, "'%s': %s" % (term, expl))


def command_add(bot, user, channel, args):
    """Adds explanation for term. Usage: .add <term> <explanation> See also: .expl, .rexpl, .del, .ls"""

    try:
        expldir, termlist = check_params(bot, args, channel)
    except TypeError:
        bot.log("No expldir for channel %s" % channel)
        return

    args = args.split(" ", 1)
    if not args[1]:
        return bot.say(user, "No explanation given")

    term = expl_parseterm(args[0])
    if term in termlist:
        return bot.say(user, "Term '%s' already exists." % term)

    expl = args[1]
    f = file(os.path.join(expldir, term), 'w')
    f.write(expl)
    f.write("\n")  # add a newline to make it easier to admin
    f.close()

    bot.log("Term '%s' for %s added by %s: %s" % (term, channel, user, term))
    return bot.say(user, "Term '%s' added: %s" % (term, expl))


def command_del(bot, user, channel, args):
    """Deletes term from explanation database. Usage: .del <term> See also: .expl, .rexpl, .add, .ls"""

    if not isAdmin(user):
        return

    try:
        expldir, termlist = check_params(bot, args, channel)
    except TypeError:
        return

    term = expl_parseterm(args)
    if not term in termlist:
        return bot.say(user, "Term '%s' doesn't exist." % term)

    expl = expl_getexpl(expldir, term)
    os.unlink(os.path.join(expldir, term))
    bot.log("Term '%s' for %s deleted by %s (contained: %s)" % (term, channel, user, expl))
    return bot.say(user, "Term '%s' deleted (contained: %s)" % (term, expl))


def command_ls(bot, user, channel, args):
    """Lists commands matching your Unix-like search query. Usage: .ls <query> Example: .ls ex[cp]la[!i]* matches 'explanation' and 'exclamation' but not 'explain'. See also: .expl, .rexpl, .add, .del"""

    try:
        expldir, termlist = check_params(bot, args, channel)
    except TypeError:
        return

    pattern = args.strip()
    matchlist = fnmatch.filter(termlist, pattern)
    matches = len(matchlist)

    if matches == 0:
        return bot.say(user, "No term matched '%s'." % pattern)
    first = ""
    if matches > 20:
        first = "first 20 of "

    matchlist.sort()
    return bot.say(user, "Terms matching '%s' (%stotal %d): %s" % (pattern, first, matches, ", ".join(matchlist[0:20])))



"""
$Id: module_imdb.py 331 2012-11-13 10:18:18Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/module_imdb.py $
"""

has_imdb = False

try:
    from imdb import IMDb
    has_imdb = True
except:
    print "Could not find IMDbPY library, please install from http://imdbpy.sourceforge.net/"
import re


def handle_url(bot, user, channel, url, msg):
    """Handle IMDB urls"""
    if not has_imdb:
        return
    m = re.match("http://.*?\.imdb\.com/title/tt([0-9]+)/?", url)
    if not m:
        return

    i = IMDb()
    movie = i.get_movie(m.group(1))
    title = movie['long imdb title']
    rating = movie.get('rating', 0.0)
    votes = movie.get('votes', 'no')
    toprank = movie.get('top 250 rank')
    rank = ""
    if toprank:
        rank = "Top 250: #%d" % toprank
    bottomrank = movie.get('bottom 100 rank')
    if bottomrank:
        rank = "Bottom 100: #%d" % bottomrank

    genre = "(" + "/".join(movie.get('genres')) + ")"

    msg = "[IMDB] %s - Rating: %.1f (%s votes) %s %s" % (title, rating, votes, genre, rank)
    msg = msg.encode("UTF-8")

    return bot.say(channel, msg)



import requests
import urllib
import logging

try:
  from lxml import etree
  print("running with lxml.etree")
except ImportError:
  print("module_wolfram_alpha requires lxml.etree for xpath support")

appid = None
query = "http://api.wolframalpha.com/v2/query?input=%s&appid=%s"

log = logging.getLogger('wolfram_alpha')

def init(bot):
    global appid
    config = bot.config.get("module_wolfram_alpha", {})
    appid = config.get("appid", "")
    if appid:
        log.info("Using Wolfram Alpha appid %s" % appid)
    else:
        log.warning("Appid not found from config!")

def command_wa(bot, user, channel, args):
    """Query Wolfram Alpha"""
    if not appid:
        log.warn("Appid not specified in configuration!")
        return

    r = requests.get(query % (urllib.quote(args), appid))

    if r.status_code != 200: return

    root = etree.fromstring(r.content)
    # find all pods
    pods = root.findall("pod")

    # no answer pods found, check if there are didyoumeans-elements
    if not pods: 
        didyoumeans = root.find("didyoumeans")
        # no support for future stuff yet, TODO?
        if not didyoumeans: return

        options = []
        for didyoumean in didyoumeans:
            options.append("'%s'" % didyoumean.text)
        line = " or ".join(options)
        line = "Did you mean %s?" % line
        return bot.say(channel, line.encode("UTF-8"))

    # first pod has the question as WA sees it
    question = pods[0].xpath("subpod/plaintext")[0].text
    # second one has the best answer
    answer = pods[1].xpath("subpod/plaintext")[0].text

    line = "%s: %s" % (question, answer)
    return bot.say(channel, line.encode("UTF-8"))



"""
$Id: module_sqliteseries.py 331 2012-11-13 10:18:18Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/available/module_sqliteseries.py $
"""

try:
    import sqlite3
except:
    import sqlite
import time
import datetime

# create table series (id INTEGER PRIMARY KEY, uid TEXT UNIQUE, serie TEXT, season INTEGER, episode INTEGER, title TEXT, airdate DATE);
# insert into series values(null, 'Alias', 1, 1, 'Pilot', date('2006-01-01'));
# select * from series where airdate = date('2006-01-02', '-1 day');
# select * from series where airdate = date('now', '-1 day');


def command_epinfo(bot, user, channel, args):
    """List series in the database"""
    # TODO: Hard-coded SHRIKE
    con = sqlite.connect("/home/shrike/pyfibot/modules/series.db")
    cur = con.cursor()
    cur.execute("SELECT DISTINCT serie FROM series")
    res = []
    for serie in cur:
        res.append(serie[0])
    res.sort()
    bot.say(channel, "Known series (" + str(len(res)) + "): " + ", ".join(res))
    cur.close()
    con.close()


def command_ep(bot, user, channel, args):
    """Usage: sqlep [today|yesterday|tomorrow] or [seriename]"""
    con = sqlite.connect("/home/shrike/pyfibot/modules/series.db")
    cur = con.cursor()

    if not args:
        bot.say(channel, "Usage: ep [today|yesterday|tomorrow] or [name of series]")
        return

    if args == "today":
        cur.execute("SELECT * FROM series WHERE airdate = date('now', 'localtime');")
        if cur.rowcount == 0:
            bot.say(channel, "No known releases today")
            return
    elif args == "yesterday":
        cur.execute("SELECT * FROM series WHERE airdate = date('now', 'localtime', '-1 day');")
        if cur.rowcount == 0:
            bot.say(channel, "No known releases yesterday")
            return
    elif args == "tomorrow":
        cur.execute("SELECT * FROM series WHERE airdate = date('now', 'localtime', '+1 day');")
        if cur.rowcount == 0:
            bot.say(channel, "No known releases tomorrow")
            return
    else:
        # try to find the serie
        cur.execute("SELECT * FROM series WHERE serie LIKE %s AND airdate >= date('now', 'localtime') LIMIT 1", ("%" + args + "%",))
        # nothing found, get more data from the web
        if cur.rowcount == 0:
            cur.execute("SELECT * FROM series WHERE serie LIKE %s", ("%" + args + "%",))
            if cur.rowcount == 0:
                # TODO: add to 'wishlist' file or something?
                bot.say(channel, "Series '%s' not found" % args)
                return
            else:
                bot.say(channel, "No unaired episodes of '%s' found" % args)
                return

    episodes = []
    # go through the results
    for (idno, uid, serie, season, episode, title, airdate) in cur:
        if episode < 10:
            episode = "0%d" % episode  # pad ep with zeroes
        # YYYY-MM-DD -> datetime -> timedelta
        t = time.strptime(airdate, "%Y-%m-%d")
        ad = datetime.date(t.tm_year, t.tm_mon, t.tm_mday)
        now = datetime.date.today()
        tomorrow = now + datetime.timedelta(days=1)
        td = ad - now

        # change 0 and 1 to today & tomorrow, don't show date if we're asking stuff for a certain day
        airdatestr = ""
        if td.days >= 0:
            if ad == now:
                if args != "today":
                    airdatestr = "on %s (Today)" % airdate
            elif ad == tomorrow:
                if args != "tomorrow":
                    airdatestr = "on %s (Tomorrow)" % airdate
            else:
                airdatestr = "on %s (%d days)" % (airdate, td.days)
        episodes.append("%s %sx%s '%s' %s" % (serie, season, episode, title, airdatestr))
    bot.say(channel, "-- ".join(episodes))
    cur.close()
    con.close()



"""
$Id: module_sqlitewanha.py 331 2012-11-13 10:18:18Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/available/module_sqlitewanha.py $
"""

try:
    import sqlite3
    sqlite = sqlite3
except:
    import sqlite
import time
import datetime
import sys
import os.path

# create table urls (id TEXT UNIQUE, nick TEXT, url TEXT, channel TEXT, time int);


def init(bot):
    global config
    config = bot.config.get("module_sqlitewanha", None)


def handle_url(bot, user, channel, url, msg):
    if not config:
        return
    ret = None
    urlid = "%s|%s" % (channel, url)
    con = sqlite.connect(os.path.join(sys.path[0], "urls.sqlite"))
    cur = con.cursor()
    cur.execute("SELECT * FROM urls WHERE id=%s", (urlid,))
    if cur.rowcount:
        id, userhost, url, channel, timestamp = cur.fetchone()
        pastetime = datetime.datetime.fromtimestamp(timestamp)
        now = datetime.datetime.now()
        age = now - pastetime
        agestr = ""
        if age.days > 0:
            agestr += "%d days " % age.days
        secs = age.seconds
        hours, minutes, seconds = secs // 3600, secs // 60 % 60, secs % 60
        if hours > 0:
            agestr += "%d h " % hours
        if minutes > 0:
            agestr += "%d m " % minutes
        if seconds > 0:
            agestr += "%d s" % seconds
        # don't alert for the same person
        if getNick(user) != getNick(userhost):
            if channel != "#wow":
                ret = bot.say(channel, "%s: wanha. (by %s %s ago)" % (getNick(user), getNick(userhost), agestr))
    else:
        cur.execute("INSERT INTO urls VALUES(%s, %s, %s, %s, %d)", (urlid, user, url, channel, int(time.time())))
    con.commit()
    cur.close()
    con.close()
    return ret



# -*- coding: utf-8 -*-
"""Displays HTML page titles

Smart title functionality for sites which could have clear titles,
but still decide show idiotic bulk data in the HTML title element

$Id: module_urltitle.py 340 2012-11-14 22:16:45Z riku.lindblad@gmail.com $
$HeadURL: https://pyfibot.googlecode.com/svn/trunk/pyfibot/modules/module_urltitle.py $
"""

import fnmatch
import urlparse
import logging
import re
try:
    import json
    has_json = True
except:
    print "Unable to load json, not all title features will work"

from types import TupleType

from BeautifulSoup import BeautifulStoneSoup

log = logging.getLogger("urltitle")
config = None


def init(bot):
    global config
    config = bot.config.get("module_urltitle", {})


def handle_url(bot, user, channel, url, msg):
    """Handle urls"""

    if msg.startswith("-"):
        return
    if re.match("http://.*?\.imdb\.com/title/tt([0-9]+)/?", url):
        return  # IMDB urls are handled elsewhere
    if re.match("(http:\/\/open.spotify.com\/|spotify:)(album|artist|track)([:\/])([a-zA-Z0-9]+)\/?", url):
        return  # spotify handled elsewhere

    if channel.lstrip("#") in config.get('disable', ''):
        return

    # hack, support both ignore and ignore_urls for a while
    for ignore in config.get("ignore", []):
        if fnmatch.fnmatch(url, ignore):
            log.info("Ignored URL: %s %s", url, ignore)
            return
    for ignore in config.get("ignore_urls", []):
        if fnmatch.fnmatch(url, ignore):
            log.info("Ignored URL: %s %s", url, ignore)
            return
    for ignore in config.get("ignore_users", []):
        if fnmatch.fnmatch(user, ignore):
            log.info("Ignored url from user: %s, %s %s", user, url, ignore)
            return

    # a crude way to handle the new-fangled shebang urls as per
    # http://code.google.com/web/ajaxcrawling/docs/getting-started.html
    # this can manage twitter + gawker sites for now
    url = url.replace("#!", "?_escaped_fragment_=")

    handlers = [(h, ref) for h, ref in globals().items() if h.startswith("_handle_")]

    # try to find a specific handler for the URL
    for handler, ref in handlers:
        pattern = ref.__doc__.split()[0]
        if fnmatch.fnmatch(url, pattern):
            title = ref(url)
            if title:
                # handler found, abort
                return _title(bot, channel, title, True)

    bs = getUrl(url).getBS()
    if not bs:
        log.debug("No BS available, returning")
        return

    title = bs.first('title')
    # no title attribute
    if not title:
        log.debug("No title found, returning")
        return

    try:
        # remove trailing spaces, newlines, linefeeds and tabs
        title = title.string.strip()
        title = title.replace("\n", " ")
        title = title.replace("\r", " ")
        title = title.replace("\t", " ")

        # compress multiple spaces into one
        title = re.sub("[ ]{2,}", " ", title)

        # nothing left in title (only spaces, newlines and linefeeds)
        if not title:
            return

        if _check_redundant(url, title):
            log.debug("Redundant title, not displaying")
            return

        return _title(bot, channel, title)

    except AttributeError:
        # TODO: Nees a better way to handle this. Happens with empty <title> tags
        pass


def _check_redundant(url, title):
    """Returns true if the url and title are similar enough."""
    # Remove hostname from the title
    hostname = urlparse.urlparse(url.lower()).netloc
    hostname = ".".join(hostname.split('@')[-1].split(':')[0].lstrip('www.').split('.'))
    cmp_title = title.lower()
    for part in hostname.split('.'):
        idx = cmp_title.replace(' ', '').find(part)
        if idx != -1:
            break

    if idx > len(cmp_title) / 2:
        cmp_title = cmp_title[0:idx + (len(title[0:idx]) - len(title[0:idx].replace(' ', '')))].strip()
    elif idx == 0:
        cmp_title = cmp_title[idx + len(hostname):].strip()
    # Truncate some nordic letters
    unicode_to_ascii = {u'\u00E4': 'a', u'\u00C4': 'A', u'\u00F6': 'o', u'\u00D6': 'O', u'\u00C5': 'A', u'\u00E5': 'a'}
    for i in unicode_to_ascii:
        cmp_title = cmp_title.replace(i, unicode_to_ascii[i])

    cmp_url = url.replace("-", " ")
    cmp_url = url.replace("+", " ")
    cmp_url = url.replace("_", " ")

    parts = cmp_url.lower().rsplit("/")

    distances = []
    for part in parts:
        if part.rfind('.') != -1:
            part = part[:part.rfind('.')]
        distances.append(_levenshtein_distance(part, cmp_title))

    if len(title) < 20 and min(distances) < 5:
        return True
    elif len(title) >= 20 and len(title) <= 30 and min(distances) < 10:
        return True
    elif len(title) > 30 and len(title) <= 60 and min(distances) <= 21:
        return True
    elif len(title) > 60 and min(distances) < 37:
        return True
    return False


def _levenshtein_distance(s, t):
    d = [[i] + [0] * len(t) for i in xrange(0, len(s) + 1)]
    d[0] = [i for i in xrange(0, (len(t) + 1))]

    for i in xrange(1, len(d)):
        for j in xrange(1, len(d[i])):
            if len(s) > i - 1 and len(t) > j - 1 and s[i - 1] == t[j - 1]:
                d[i][j] = d[i - 1][j - 1]
            else:
                d[i][j] = min((d[i - 1][j] + 1, d[i][j - 1] + 1, d[i - 1][j - 1] + 1))

    return d[len(s)][len(t)]


def _title(bot, channel, title, smart=False, prefix=None):
    """Say title to channel"""

    if not title: return

    if not prefix:
        prefix = "Title:"
    info = None
    # tuple, additional info
    if type(title) == TupleType:
        info = title[1]
        title = title[0]
    # crop obscenely long titles
    if len(title) > 200:
        title = title[:200] + "..."

    title = BeautifulStoneSoup(title, convertEntities=BeautifulStoneSoup.ALL_ENTITIES)
    log.info(title)

    if not info:
        return bot.say(channel, "%s '%s'" % (prefix, title))
    else:
        return bot.say(channel, "%s '%s' [%s]" % (prefix, title, info))


# TODO: Some handlers does not have if not bs: return, but why do we even have this for every function
def _handle_iltalehti(url):
    """*iltalehti.fi*html"""
    # Go as normal
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.first('title').string
    # The first part is the actual story title, lose the rest
    title = title.split("|")[0].strip()
    return title


def _handle_iltasanomat(url):
    """*iltasanomat.fi*uutinen.asp*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.title.string.rsplit(" - ", 1)[0]
    return title


def _handle_keskisuomalainen_sahke(url):
    """*keskisuomalainen.net*sahkeuutiset/*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.first('p', {'class': 'jotsikko'})
    if title:
        title = title.next.strip()
        return title


def _handle_tietokone(url):
    """http://www.tietokone.fi/uutta/uutinen.asp?news_id=*"""
    bs = getUrl(url).getBS()
    sub = bs.first('h5').string
    main = bs.first('h2').string
    return "%s - %s" % (main, sub)


def _handle_itviikko(url):
    """http://www.itviikko.fi/*/*/*/*/*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    return bs.first("h1", "headline").string

def _handle_verkkokauppa(url):
    """http://www.verkkokauppa.com/*/product/*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    product = bs.first('h1', id='productName').string
    price = bs.first('span', {'class': 'hintabig'}).string
    return "%s | %s" % (product, price)


def _handle_mol(url):
    """http://www.mol.fi/paikat/Job.do?*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.first("div", {'class': 'otsikko'}).string
    return title

def _handle_tweet2(url):
    """http*://twitter.com/*/status/*"""
    return _handle_tweet(url)

def _handle_tweet(url):
    """http*://twitter.com/*/statuses/*"""
    import simplejson as json
    import urllib2
    tweet_url = "http://api.twitter.com/1/statuses/show/%s.json"
    test = re.match("https?://twitter\.com\/(\w+)/status(es)?/(\d+)",url)
    #    matches for unique tweet id string
    infourl = tweet_url % test.group(3)

    twitapi = urllib2.urlopen(infourl)
    #loads into dict
    json1 = json.load(twitapi)

    #reads dict
    ##You can modify the fields below or add any fields you want to the returned string
    text = json1['text']
    user = json1['user']['screen_name']
    name = json1['user']['name']
    tweet = "Tweet by %s(@%s): %s" % (name, user, text)
    return tweet


def _handle_netanttila(url):
    """http://www.netanttila.com/webapp/wcs/stores/servlet/ProductDisplay*"""
    bs = getUrl(url).getBS()
    itemname = bs.first("h1").string.replace("\n", "").replace("\r", "").replace("\t", "").strip()
    price = bs.first("td", {'class': 'right highlight'}).string.split(" ")[0]
    return "%s | %s EUR" % (itemname, price)


def _handle_youtube_shorturl(url):
    """http*://youtu.be/*"""
    return _handle_youtube_gdata(url)


def _handle_youtube_gdata_new(url):
    """http*://youtube.com/watch#!v=*"""
    return _handle_youtube_gdata(url)


def _handle_youtube_gdata(url):
    """http*://*youtube.com/watch?*v=*"""
    gdata_url = "http://gdata.youtube.com/feeds/api/videos/%s"

    match = re.match("https?://youtu.be/(.*)", url)
    if not match:
        match = re.match("https?://.*?youtube.com/watch\?.*?v=([^&]+)", url)
    if match:
        infourl = gdata_url % match.group(1)
        bs = getUrl(infourl, True).getBS()

        entry = bs.first("entry")

        if not entry:
            log.info("Video too recent, no info through API yet.")
            return

        author = entry.author.next.string
        # if an entry doesn't have a rating, the whole element is missing
        try:
            rating = float(entry.first("gd:rating")['average'])
        except TypeError:
            rating = 0.0

        stars = int(round(rating)) * "*"
        statistics = entry.first("yt:statistics")
        if statistics:
            views = statistics['viewcount']
        else:
            views = "no"
        racy = entry.first("yt:racy")
        media = entry.first("media:group")
        title = media.first("media:title").string
        secs = int(media.first("yt:duration")['seconds'])
        lengthstr = []
        hours, minutes, seconds = secs // 3600, secs // 60 % 60, secs % 60
        if hours > 0:
            lengthstr.append("%dh" % hours)
        if minutes > 0:
            lengthstr.append("%dm" % minutes)
        if seconds > 0:
            lengthstr.append("%ds" % seconds)
        if racy:
            adult = " - XXX"
        else:
            adult = ""
        return "%s by %s [%s - %s - %s views%s]" % (title, author, "".join(lengthstr), "[%-5s]" % stars, views, adult)


def _handle_helmet(url):
    """http://www.helmet.fi/record=*fin"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.find(attr={'class': 'bibInfoLabel'}, text='Teoksen nimi').next.next.next.next.string
    return title


def _handle_ircquotes(url):
    """http://*ircquotes.fi/[?]*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    chan = bs.first("span", {'class': 'quotetitle'}).next.next.string
    points = bs.first("span", {'class': 'points'}).next.string
    firstline = bs.first("div", {'class': 'quote'}).next.string
    title = "%s (%s): %s" % (chan, points, firstline)
    return title


def _handle_alko2(url):
    """http://alko.fi/tuotteet/fi/*"""
    return _handle_alko(url)


def _handle_alko(url):
    """http://www.alko.fi/tuotteet/fi/*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    name = bs.find('span', {'class': 'tuote_otsikko'}).string
    price = bs.find('span', {'class': 'tuote_hinta'}).string.split(" ")[0] + u"€"
    drinktype = bs.find('span', {'class': 'tuote_tyyppi'}).next
    return name + " - " + drinktype + " - " + price


def _handle_salakuunneltua(url):
    """*salakuunneltua.fi*"""
    return None


def _handle_facebook(url):
    """*facebook.com/*"""
    if not has_json: return
    if re.match("http(s?)://(.*?)facebook\.com/(.*?)id=(\\d+)", url):
        asd = urlparse.urlparse(url)
        id = asd.query.split('id=')[1].split('&')[0]
        if id != '':
            url = "https://graph.facebook.com/%s" % id
            content = getUrl(url, True).getContent()
            if content != 'false':
                data = json.loads(content)
                try:
                    title = data['name']
                except:
                    return
            else:
                title = 'Private url'
    else:
        return
    return title


def _handle_vimeo(url):
    """*vimeo.com/*"""
    data_url = "http://vimeo.com/api/v2/video/%s.xml"
    match = re.match("http://.*?vimeo.com/(\d+)", url)
    if match:
        infourl = data_url % match.group(1)
        bs = getUrl(infourl, True).getBS()
        title = bs.first("title").string
        user = bs.first("user_name").string
        likes = bs.first("stats_number_of_likes").string
        plays = bs.first("stats_number_of_plays").string
        return "%s by %s [%s likes, %s views]" % (title, user, likes, plays)


def _handle_stackoverflow(url):
    """*stackoverflow.com/questions/*"""
    if not has_json: return
    api_url = 'http://api.stackoverflow.com/1.1/questions/%s'
    match = re.match('.*stackoverflow.com/questions/([0-9]+)', url)
    if match is None:
        return
    question_id = match.group(1)
    content = getUrl(api_url % question_id, True).getContent()
    if not content:
        log.debug("No content received")
        return
    try:
        data = json.loads(content)
        title = data['questions'][0]['title']
        tags = "/".join(data['questions'][0]['tags'])
        score = data['questions'][0]['score']
        return "%s - %dpts - %s" % (title, score, tags)
    except Exception, e:
        return "Json parsing failed %s" % e


def _handle_hs(url):
    """*hs.fi*artikkeli*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.title.string
    title = title.split("-")[0].strip()
    try:
        # determine article age and warn if it is too old
        from datetime import datetime
        # handle updated news items of format, and get the latest update stamp
        # 20.7.2010 8:02 | PÃ¤ivitetty: 20.7.2010 12:53
        date = bs.first('p', {'class': 'date'}).next
        # in case hs.fi changes the date format, don't crash on it
        if date:
            date = date.split("|")[0].strip()
            article_date = datetime.strptime(date, "%d.%m.%Y %H:%M")
            delta = datetime.now() - article_date

            if delta.days > 365:
                return title, "NOTE: Article is %d days old!" % delta.days
            else:
                return title
        else:
            return title
    except Exception, e:
        log.error("Error when parsing hs.fi: %s" % e)
        return title

def _handle_mtv3(url):
    """*mtv3.fi*"""
    bs = getUrl(url).getBS()
    title = bs.first("h1", "otsikko").next
    return title


def _handle_yle(url):
    """http://*yle.fi/uutiset/*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.title.string
    title = title.split("|")[0].strip()
    return title


def _handle_varttifi(url):
    """http://www.vartti.fi/artikkeli/*"""
    bs = getUrl(url).getBS()
    title = bs.first("h2").string
    return title


def _handle_aamulehti(url):
    """http://www.aamulehti.fi/*"""
    bs = getUrl(url).getBS()
    if not bs:
        return
    title = bs.fetch("h1")[0].string
    return title


def _handle_apina(url):
    """http://apina.biz/*"""
    return None
